// Copyright (c) 2022 Timofey Solomko
// Licensed under MIT License
//
// See LICENSE for license information

import XCTest
import BitByteData

class MsbBitReaderBenchmarks: XCTestCase {

    func testAdvance() {
        self.measure {
            let reader = MsbBitReader(data: Data(count: 10_485_760)) // 10 MB

            for _ in 0..<5_000_000 * 8 {
                reader.advance()
            }
        }
    }

    func testAdvanceRealistic() {
        self.measure {
            let reader = MsbBitReader(data: Data(count: 10_485_760)) // 10 MB

            for _ in 0..<9_300_000 {
                reader.advance(by: 6)
                reader.advance(by: 3)
            }
        }
    }

    func testBit() {
        self.measure {
            let reader = MsbBitReader(data: Data(count: 10_485_760)) // 10 MB

            for _ in 0..<5_000_000 * 8 {
                _ = reader.bit()
            }
        }
    }

    func testBits() {
        self.measure {
            let reader = MsbBitReader(data: Data(count: 10_485_760)) // 10 MB

            for _ in 0..<1_000_000 * 8 {
                _ = reader.bits(count: 5)
            }
        }
    }

    func testIntFromBits() {
        self.measure {
            let reader = MsbBitReader(data: Data(count: 10_485_760)) // 10 MB

            for _ in 0..<1_000_000 * 4 {
                _ = reader.int(fromBits: 10)
            }
        }
    }

    func testSignedInt_SM_pos() {
        var bytes = [UInt8]()
        for _ in 0..<5_242_880 { // 5 MB * 2
            bytes.append(0xD)
            bytes.append(0x37)
        }

        self.measure {
            let reader = MsbBitReader(data: Data(bytes))

            for _ in 0..<5_000_000 {
                _ = reader.signedInt(fromBits: 16, representation: .signMagnitude)
            }
        }
    }

    func testSignedInt_SM_neg() {
        var bytes = [UInt8]()
        for _ in 0..<5_242_880 { // 5 MB * 2
            bytes.append(0x8D)
            bytes.append(0x37)
        }

        self.measure {
            let reader = MsbBitReader(data: Data(bytes))

            for _ in 0..<5_000_000 {
                _ = reader.signedInt(fromBits: 16, representation: .signMagnitude)
            }
        }
    }

    func testSignedInt_1C_pos() {
        var bytes = [UInt8]()
        for _ in 0..<5_242_880 { // 5 MB * 2
            bytes.append(0xD)
            bytes.append(0x37)
        }

        self.measure {
            let reader = MsbBitReader(data: Data(bytes))

            for _ in 0..<5_000_000 {
                _ = reader.signedInt(fromBits: 16, representation: .oneComplementNegatives)
            }
        }
    }

    func testSignedInt_1C_neg() {
        var bytes = [UInt8]()
        for _ in 0..<5_242_880 { // 5 MB * 2
            bytes.append(0x8D)
            bytes.append(0x37)
        }

        self.measure {
            let reader = MsbBitReader(data: Data(bytes))

            for _ in 0..<5_000_000 {
                _ = reader.signedInt(fromBits: 16, representation: .oneComplementNegatives)
            }
        }
    }

    func testSignedInt_2C_pos() {
        var bytes = [UInt8]()
        for _ in 0..<5_242_880 { // 5 MB * 2
            bytes.append(0xD)
            bytes.append(0x37)
        }

        self.measure {
            let reader = MsbBitReader(data: Data(bytes))

            for _ in 0..<5_000_000 {
                _ = reader.signedInt(fromBits: 16, representation: .twoComplementNegatives)
            }
        }
    }

    func testSignedInt_2C_neg() {
        var bytes = [UInt8]()
        for _ in 0..<5_242_880 { // 5 MB * 2
            bytes.append(0x8D)
            bytes.append(0x37)
        }

        self.measure {
            let reader = MsbBitReader(data: Data(bytes))

            for _ in 0..<5_000_000 {
                _ = reader.signedInt(fromBits: 16, representation: .twoComplementNegatives)
            }
        }
    }

    func testSignedInt_E127() {
        var bytes = [UInt8]()
        for _ in 0..<5_242_880 { // 5 MB * 2
            bytes.append(0x6D)
            bytes.append(0xB7)
        }

        self.measure {
            let reader = MsbBitReader(data: Data(bytes))

            for _ in 0..<5_000_000 {
                _ = reader.signedInt(fromBits: 7, representation: .biased(bias: 127))
            }
        }
    }

    func testSignedInt_RN2() {
        var bytes = [UInt8]()
        for _ in 0..<5_242_880 { // 5 MB * 2
            bytes.append(0x6D)
            bytes.append(0xB7)
        }

        self.measure {
            let reader = MsbBitReader(data: Data(bytes))

            for _ in 0..<5_000_000 {
                _ = reader.signedInt(fromBits: 13, representation: .radixNegativeTwo)
            }
        }
    }

    func testByteFromBits() {
        self.measure {
            let reader = MsbBitReader(data: Data(count: 10_485_760)) // 10 MB

            for _ in 0..<1_000_000 * 8 {
                _ = reader.byte(fromBits: 6)
            }
        }
    }

    func testUint16FromBits() {
        self.measure {
            let reader = MsbBitReader(data: Data(count: 10_485_760)) // 10 MB

            for _ in 0..<1_000_000 * 4 {
                _ = reader.uint16(fromBits: 13)
            }
        }
    }

    func testUint32FromBits() {
        self.measure {
            let reader = MsbBitReader(data: Data(count: 10_485_760)) // 10 MB

            for _ in 0..<1_000_000 * 3 {
                _ = reader.uint32(fromBits: 23)
            }
        }
    }

    func testUint64FromBits() {
        self.measure {
            let reader = MsbBitReader(data: Data(count: 10_485_760)) // 10 MB

            for _ in 0..<1_000_000 {
                _ = reader.uint64(fromBits: 52)
            }
        }
    }

}
